@@ -129,14 +129,7 @@ def check_git_repo(args):
     """
     return
 
-def get_leafs_that_need_to_reproduce():
-    from dvc.repo import Repo as DVCRepo
-    dvcrepo = DVCRepo('.')
-    try:
-        Gs = dvcrepo.pipelines
-    except:
-        # OLDER DVC-VERSION !
-        Gs = dvcrepo.pipelines()
+def get_leafs_that_need_to_reproduce(dvcrepo, Gs):
     # Get all leaf files.
     leafs = []
     for G in Gs:
@@ -170,25 +163,17 @@ def get_leafs_that_need_to_reproduce():
             leafs_need_to_reproduce.append(leafs[i])
     return leafs_need_to_reproduce
 
-def all_jupyter_notebook_to_py_files(project_dir):
+def all_jupyter_notebook_to_py_files(Gs):
     created_files = []
-
-    from dvc.repo import Repo as DVCRepo
-    dvcrepo = DVCRepo('.')
-    try:
-        Gs = dvcrepo.pipelines
-    except:
-        # OLDER DVC-VERSION !
-        Gs = dvcrepo.pipelines()
-
     for G in Gs:
         for s in G.nodes():
-            cmd = s.cmd().split()
+            cmd = s.cmd.split()
             if cmd[0] == 'python':
                 for c in cmd:
                     if c.endswith('.py'):
                         if not os.path.exists(c) and os.path.exists(c[:-3]+'.ipynb'):
-                            created_files.append(jupyter_notebook_to_py_file(c))
+                            created_files.append(jupyter_notebook_to_py_file(c[:-3]+'.ipynb'))
+                            print(bcolors.BOLD+'    Convert ' + c[:-3]+'.ipynb' +  ' to ' + c + ' file.' + bcolors.ENDC)
     return created_files
 
 def jupyter_notebook_to_py_file(path_to_ipynb):
@@ -436,6 +421,54 @@ def define_the_exp_name(exp_name, hyperopt_draws, list_of_variables):
 
     return batch_names
 
+def find_all_dvc_leafs(args_dvc_files):
+
+    #####################################
+    # Rename the hyperopt-files to .dvc #
+    #####################################
+    if os.path.exists('dvc') and os.path.exists(str(Path('dvc/.hyperopt'))):
+        list_of_hyperopt_files = [f for f in os.listdir(str(Path('dvc/.hyperopt'))) if f.endswith('.hyperopt')]
+        for f in list_of_hyperopt_files:
+            os.rename(str(Path('dvc/.hyperopt/' + f)), str(Path('dvc/'+f[:-9]+'.dvc')))
+            print(str(Path('dvc/.hyperopt/' + f)), str(Path('dvc/'+f[:-9]+'.dvc')))
+    else:
+        list_of_hyperopt_files = []
+
+    #############################
+    # Find all leafs to execute #
+    #############################
+    try:
+        dvcrepo = DVCRepo('.')
+        Gs = dvcrepo.pipelines
+
+        if len(Gs) > 0:
+            if args_dvc_files is None:
+                #dvc_files = [[f[2:]] for f in helper.getListOfFiles(add_only_files_that_ends_with='.dvc')]
+                dvc_files = get_leafs_that_need_to_reproduce(dvcrepo, Gs)
+            else:
+                dvc_files = []
+                dvc_files_tmp = args_dvc_files.dvc_files.replace(' ', '').split(',')
+                for dvc_files_branch in dvc_files_tmp:
+                    dvc_files.append([])
+                    dvc_files_file = dvc_files_branch.split('|')
+                    for i in range(len(dvc_files_file)):
+                        dvc_files[-1].append(dvc_files_file[i])
+                        if not dvc_files_file[i].endswith('.dvc'):
+                            raise ValueError('Error: You define with -f which dvc files you want to exec. One or more files does not ends with .dvc. Please use only DVC files.')
+        else:
+            dvc_files = []
+    # make sure that the hyperopt files get always renamed!
+    finally:
+        #############################
+        # Rename the hyperopt-files #
+        #############################
+        for f in list_of_hyperopt_files:
+            if os.path.exists(str(Path('dvc/'+f[:-9]+'.dvc'))):
+                os.rename(str(Path('dvc/'+f[:-9]+'.dvc')), str(Path('dvc/.hyperopt/' + f)))
+            else:
+                print(bcolors.WARNING+'Warning: File ' + str(Path('dvc/'+f[:-9]+'.dvc')) + ' not found.'+bcolors.ENDC)
+    return dvc_files, list_of_hyperopt_files, Gs
+
 def main():
 
     parser = ArgumentParser(description=DESCRIPTION)
@@ -459,6 +492,9 @@ def main():
                              'https://www.curious-containers.cc/docs/red-format-protecting-credentials',
                         default= None)
 
+    parser.add_argument('-p','--papermill',help='Use papermill to run the jupyter notebook on the server and save the results in the jupyter notebook. If this parameter is set, no jupyter notebook will be converted to py files.',
+                        default=False, action='store_true')
+
     args = parser.parse_args()
     
     project_dir = get_main_git_directory_Path()
@@ -507,76 +543,19 @@ def main():
     except:
         print('Some files are missing.')
 
-    #####################################
-    # Rename the hyperopt-files to .dvc #
-    #####################################
-    if os.path.exists('dvc') and os.path.exists(str(Path('dvc/.hyperopt'))):
-        list_of_hyperopt_files = [f for f in os.listdir(str(Path('dvc/.hyperopt'))) if f.endswith('.hyperopt')]
-        for f in list_of_hyperopt_files:
-            os.rename(str(Path('dvc/.hyperopt/' + f)), str(Path('dvc/'+f[:-9]+'.dvc')))
-            print(str(Path('dvc/.hyperopt/' + f)), str(Path('dvc/'+f[:-9]+'.dvc')))
-    else:
-        list_of_hyperopt_files = []
-
-    #############################
-    # Find all leafs to execute #
-    #############################
-    try:
-        if args.dvc_files is None:
-            #dvc_files = [[f[2:]] for f in helper.getListOfFiles(add_only_files_that_ends_with='.dvc')]
-            dvc_files = get_leafs_that_need_to_reproduce()
-        else:
-            dvc_files = []
-            dvc_files_tmp = args.dvc_files.replace(' ', '').split(',')
-            for dvc_files_branch in dvc_files_tmp:
-                dvc_files.append([])
-                dvc_files_file = dvc_files_branch.split('|')
-                for i in range(len(dvc_files_file)):
-                    dvc_files[-1].append(dvc_files_file[i])
-                    if not dvc_files_file[i].endswith('.dvc'):
-                        raise ValueError('Error: You define with -f which dvc files you want to exec. One or more files does not ends with .dvc. Please use only DVC files.')
-    # make sure that the hyperopt files get always renamed!
-    finally:
-        #############################
-        # Rename the hyperopt-files #
-        #############################
-        for f in list_of_hyperopt_files:
-            if os.path.exists(str(Path('dvc/'+f[:-9]+'.dvc'))):
-                os.rename(str(Path('dvc/'+f[:-9]+'.dvc')), str(Path('dvc/.hyperopt/' + f)))
-            else:
-                print(bcolors.WARNING+'Warning: File ' + str(Path('dvc/'+f[:-9]+'.dvc')) + ' not found.'+bcolors.ENDC)
+    #############################################################
+    # Find all hyperopt files and leafs to execute the pipeline #
+    #############################################################
+    dvc_files, list_of_hyperopt_files, Gs = find_all_dvc_leafs(args.dvc_files)
 
     ####################################
     # Error if no DVC-file was defined #
     ####################################
-    if len(dvc_files) == 0:
+    if len(dvc_files) == 0 and not args.papermill:
         raise ValueError('There exist no job to execute! Create DVC-Files with "dvc run --no-exec ..." to define the jobs. Or check the .dvc_cc/dvc_cc_ignore file. All DVC-Files that are defined there are ignored from this script.')
 
-    ##########################
-    # Get All Hyperparemters #
-    ##########################
-    vc = VariableCache()
-
-    for f in list_of_hyperopt_files:
-        f = str(Path('dvc/.hyperopt/' + f))
-        vc.register_dvccc_file(f)
-
-    ###################################
-    # DEFINE ALL Hyperopt-Experiments #
-    ###################################
-    if len(vc.list_of_all_variables) > 0:
-        hyperopt_draws = create_hyperopt_variables(vc)
-        user_input = input('You defined ' + str(len(hyperopt_draws)) + ' * ' + str(args.num_of_repeats) + ' = ' + str(
-            len(
-                hyperopt_draws) * args.num_of_repeats) + ' hyperoptimization pairs. Do you want to continue and start the job? [y,n]: ')
-        if not user_input.lower().startswith('y'):
-            print('The job was canceled')
-            exit(0)
-    else:
-        hyperopt_draws = [[]]
-
     loaded_yml = None
-    cc_ids = []
+
     try:
         ###########################
         # Create an input branch! #
@@ -591,8 +570,8 @@ def main():
         #############################
         # CONVERT Jupyter Notebooks #
         #############################
-        if not args.not_ipynb_to_py:
-            created_pyfiles_from_jupyter = all_jupyter_notebook_to_py_files(project_dir)
+        if not args.not_ipynb_to_py and len(Gs) > 0:
+            created_pyfiles_from_jupyter = all_jupyter_notebook_to_py_files(Gs)
         else:
             created_pyfiles_from_jupyter = []
         for f in created_pyfiles_from_jupyter:
@@ -602,6 +581,59 @@ def main():
             subprocess.call(['git', 'commit','-q','-m', 'Convert Jupyter Notebooks to Py-File.'])
             subprocess.call(['git', 'push', '-q', '-u', 'origin', exp_name + ':' + exp_name])
 
+        ######################################
+        # Use papermill to create a dvc file #
+        ######################################
+        if args.papermill:
+            ipynb_files_in_main_dir = [f for f in os.listdir() if f.endswith('.ipynb')]  # Todo: Allow different location of the ipynb file
+            if len(ipynb_files_in_main_dir) == 0:
+                raise ValueError('To use papermill you need a jupyter notebook to run on the server in the main git '
+                                 'directory.')
+            else:
+                ipynb_files_in_main_dir = ipynb_files_in_main_dir[0]
+            print(bcolors.BOLD + 'Create a DVC file for executing ' + ipynb_files_in_main_dir + ' with papermill.' +
+                  bcolors.ENDC)
+            from dvc_cc.run import papermill_helper
+            parameters = papermill_helper.read_params_from_parametercell(ipynb_files_in_main_dir)
+
+            cmd = 'papermill ' + ipynb_files_in_main_dir + ' ' + ipynb_files_in_main_dir[:-6] + '_output.ipynb'
+            for p in parameters:
+                cmd = cmd +' -k python -p ' + p[0] + ' {{' + p[0] + ':' + p[1] + ':None}}'
+
+            subprocess.call(['dvc-cc', 'hyperopt', 'new', '-d', ipynb_files_in_main_dir, '-o',
+                             ipynb_files_in_main_dir[:-6] + '_output.ipynb', '-f','papermill.dvc',cmd])
+            subprocess.call(['git', 'add', 'dvc/.hyperopt/papermill.hyperopt'])  # TODO: build quite mode!
+            subprocess.call(['git', 'commit','-q','-m', 'Create a dvc file for papermill.'])
+            subprocess.call(['git', 'push', '-q', '-u', 'origin', exp_name + ':' + exp_name])
+            dvc_files, list_of_hyperopt_files, Gs = find_all_dvc_leafs(args.dvc_files)
+
+        ##########################
+        # Get All Hyperparemters #
+        ##########################
+        vc = VariableCache()
+
+        for f in list_of_hyperopt_files:
+            f = str(Path('dvc/.hyperopt/' + f))
+            vc.register_dvccc_file(f)
+
+        ###################################
+        # DEFINE ALL Hyperopt-Experiments #
+        ###################################
+        if len(vc.list_of_all_variables) > 0:
+            hyperopt_draws = create_hyperopt_variables(vc)
+            user_input = input(
+                'You defined ' + str(len(hyperopt_draws)) + ' * ' + str(args.num_of_repeats) + ' = ' + str(
+                    len(
+                        hyperopt_draws) * args.num_of_repeats) + ' hyperoptimization pairs. Do you want to continue and start the job? [y,n]: ')
+            if not user_input.lower().startswith('y'):
+                print('The job was canceled')
+                exit(0)
+        else:
+            hyperopt_draws = [[]]
+
+        loaded_yml = None
+
+
         #####################################################
         # TODO: SAVE the Hyperopt-Values and the VC!        #
         #   WITH THIS It is possible to get rerun the code. #
